#include <stdio.h>

#include "lnm/http/consts.h"
#include "lnm/http/loop.h"
#include "lnm/loop.h"

#include "event_loop.h"
#include "http/res.h"
#include "http/types.h"
#include "lander.h"
#include "log.h"
#include "lsm/store.h"

static const char index_page[] =
    "<!DOCTYPE html>\n"
    "<html>\n"
    "  <body>\n"
    "    <h1>r8r.be</h1>\n"
    "    <p>This is the URL shortener and pastebin accompanying my site, <a "
    "href=\"https://rustybever.be\">The Rusty Bever</a>.</p>\n"
    "  </body>\n"
    "</html>\n";

lnm_http_step_err lander_get_index(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_res_body_set_buf(&ctx->res, (char *)index_page,
                            sizeof(index_page) - 1, false);
  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_type, "text/html",
                          false);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_redirect(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  // For redirects, the URL is stored as an in-memory attribute
  lsm_str *url_attr_val;

  // This shouldn't be able to happen
  if (lsm_entry_attr_get(&url_attr_val, c_ctx->entry, lander_attr_type_url) !=
      lsm_error_ok) {
    error("Entry of type redirect detected without URL attribute");

    ctx->res.status = lnm_http_status_internal_server_error;
    lsm_entry_close(c_ctx->entry);
    c_ctx->entry = NULL;

    return lnm_http_step_err_res;
  }

  lnm_http_res_add_header_len(&ctx->res, lnm_http_header_location,
                              (char *)lsm_str_ptr(url_attr_val),
                              lsm_str_len(url_attr_val), false);

  ctx->res.status = lnm_http_status_moved_permanently;

  return lnm_http_step_err_done;
}

lnm_err lander_entry_data_streamer(uint64_t *written, char *buf,
                                   lnm_http_conn *conn, uint64_t offset,
                                   uint64_t len) {
  // TODO respect offset variable

  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  lsm_entry_data_read(written, buf, c_ctx->entry, len);

  return lnm_err_ok;
}

lnm_http_step_err lander_get_paste(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  lnm_http_res_body_set_fn(&ctx->res, lander_entry_data_streamer,
                           lsm_entry_data_len(c_ctx->entry));
  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_type, "text/plain",
                          false);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_file(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  lnm_http_res_body_set_fn(&ctx->res, lander_entry_data_streamer,
                           lsm_entry_data_len(c_ctx->entry));
  lander_attr_to_header(ctx, lander_attr_type_content_type,
                        lnm_http_header_content_type);

  lsm_str *value;
  char *buf;

  if (lsm_entry_attr_get(&value, c_ctx->entry, lander_attr_type_file_name) ==
      lsm_error_ok) {
    buf = malloc(24 + lsm_str_len(value));
    int len = lsm_str_len(value);
    sprintf(buf, "attachment; filename=\"%.*s\"", len, lsm_str_ptr(value));
  } else {
    buf = malloc(11);
    strcpy(buf, "attachment");
  }

  lnm_http_res_add_header(&ctx->res, lnm_http_header_content_disposition, buf,
                          true);

  return lnm_http_step_err_done;
}

lnm_http_step_err lander_get_entry(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;
  lnm_http_loop_gctx *gctx = ctx->g;
  lander_gctx *c_gctx = gctx->c;

  const char *key_s = &ctx->req.path.s[ctx->req.path.groups[1].rm_so];
  int key_len = ctx->req.path.groups[1].rm_eo - ctx->req.path.groups[1].rm_so;

  lsm_str *key;
  lsm_str_init_copy_n(&key, (char *)key_s, key_len);

  switch (lsm_store_open_read(&c_ctx->entry, c_gctx->store, key)) {
  case lsm_error_ok:
    break;
  case lsm_error_not_found:
    ctx->res.status = lnm_http_status_not_found;
    return lnm_http_step_err_res;
  default:
    ctx->res.status = lnm_http_status_internal_server_error;
    return lnm_http_step_err_res;
  }

  lander_entry_type t;
  lsm_entry_attr_get_uint8_t((uint8_t *)&t, c_ctx->entry,
                             lander_attr_type_entry_type);

  lnm_http_step_err res;

  switch (t) {
  case lander_entry_type_redirect:
    res = lander_get_redirect(conn);
    break;
  case lander_entry_type_paste:
    res = lander_get_paste(conn);
    break;
  case lander_entry_type_file:
    res = lander_get_file(conn);
    break;
  }

  return res;
}

bool lander_stream_body_to_client(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;
  lander_ctx *c_ctx = ctx->c;

  if ((c_ctx->entry == NULL) ||
      (ctx->res.body.expected_len == ctx->res.body.len)) {
    return true;
  }

  uint64_t to_write = MIN(EVENT_LOOP_BUFFER_SIZE - conn->wbuf_size,
                          ctx->res.body.expected_len - ctx->res.body.len);

  uint64_t read = 0;
  lsm_entry_data_read(&read, (char *)&conn->wbuf[conn->wbuf_size], c_ctx->entry,
                      to_write);

  ctx->res.body.len += read;
  conn->wbuf_size += read;

  return false;
}
